/*
 * Eucalyptus
 * by Paul A. Schifferer
 *
 * Module: compose_write.c
 *   	 Message generation functions for composing
 *
 * Copyright  1997-9 Isengard Developments
 * All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "eucalyptus.h"

gboolean compose_write_msg(Composition *comp, QueueInfo *qi)
{
	Folder *folder;
	GList *node, *to_list = NULL, *cc_list = NULL, *bcc_list = NULL;
	struct tm *tm;
	time_t t;
	glong error = 0, tos = 0, ccs = 0, bccs = 0;
	gchar *eoltypes[] = {"\r\n", "\r", "\n", };
	gchar tempfile[FILENAME_MAX], *msgfile, *addrfile, *infofile, *dir, *c;
	gboolean success = TRUE;
	FILE *fp, *addr_fp;

	if(!*uprefs.RealName ||
	   !*uprefs.Address)
	{
		prefs_user_open_window();
		euc_request(comp->Window,
				_("Eucalyptus warning"),
				_("Can't send messages without your name and\n"
				  "email address first.  Please add this information\n"
				  "and try again.  Thanks."),
				euc_ok_button, 0,
				0);
		return FALSE;
	}

	if(!(folder = folder_get_special(FTF_OUTGOING)))
	{
		if(!(comp->Flags & CF_AUTO))
			euc_request(comp->Window,
					_("Eucalyptus error"),
					_("Unable to open Outgoing folder."),
					euc_ok_button, 0,
					0);
		return FALSE;
	}

	to_list = compose_parse_addresses(comp, CAT_TO);
	cc_list = compose_parse_addresses(comp, CAT_CC);
	bcc_list = compose_parse_addresses(comp, CAT_BCC);
	
	tos = g_list_length(to_list);
	ccs = g_list_length(cc_list);
	bccs = g_list_length(bcc_list);

	if(!tos)
	{
		if(!(comp->Flags & CF_AUTO))
			euc_request(comp->Window,
					_("Eucalyptus error"),
					_("You need at least one (1) \"to\" address in\nyour list of recipients."),
					euc_ok_button, 0,
					0);
		return FALSE;
	}

	/* Get current time.
	 */
	t = time(NULL);
	tm = localtime(&t);

	/* Get next msg number for the folder
	 * (Outgoing) and set up the filenames.
	 */
	if(!folder_get_next_msg_num(folder, NULL, tempfile, sizeof(tempfile)))
		return FALSE;
	dir = folder_get_full_path(folder);
	msgfile = g_strdup(tempfile);
	c = strstr(tempfile, G_DIR_SEPARATOR_S "msg");
	memcpy(c + 1, "__a", 3);
	addrfile = g_strdup(tempfile);
	memcpy(c + 1, "__i", 3);
	infofile = g_strdup(tempfile);
	
	/* Open output file.
	 */
	if(fp = fopen(msgfile, "w"))
	{
		/* Open address list file.
		 */
		if(addr_fp = fopen(addrfile, "w"))
		{
			glong line = 0, addr_count, x, i;
			gchar *charset = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890";
			gchar *priorities[] = {"low", "normal", "high", };
			gchar addr_type = CAT_TO, *text = NULL;
			gchar *work;
			gboolean multipart = FALSE;

			/* Reset the random seed.
			 */
			srand(time(NULL));
			
			/* Generate a boundary string.
			 */
			memset(comp->Boundary, 0, sizeof(comp->Boundary));
			for(x=0; x<sizeof(comp->Boundary)-1; x++)
				comp->Boundary[x] = charset[rand() % strlen(charset)];

			/* From: header.
			 */
			work = g_strdup_printf("From: %s <%s>%s", uprefs.RealName, uprefs.Address, eoltypes[prefs.EOLType]);
			header_encode(fp, work);
			g_free(work);
			
			/* Organization: header.
			 */
			if(*uprefs.Organization)
			{
				work = g_strdup_printf("Organization: %s%s", uprefs.Organization, eoltypes[prefs.EOLType]);
				header_encode(fp, work);
				g_free(work);
			}

			// to/cc/bcc
			for(i=0; i<3; i++)
			{
				GList *addr_list;
				glong last_addr;

				switch(i)
				{
					case 0:   // to
						fprintf(fp, "To: ");
						addr_type = CAT_TO;
						addr_list = to_list;
						break;
					case 1:   // cc
						if(!ccs)
							continue;

						fprintf(fp, "Cc: ");
						addr_type = CAT_CC;
						addr_list = cc_list;
						break;
					case 2:   // bcc
						if(!bccs)
							continue;

						addr_type = CAT_BCC;
						addr_list = bcc_list;
						break;
				}

				line = 4;
				addr_count = 0;
				last_addr = 0;
				
				/* Walk the list.
				 */
				for(node = g_list_first(addr_list); node; node = g_list_next(node))
				{
					struct AddressEntry *ae;
					struct List *ae_list;
					glong sub_addr_count = 0;
					gchar *addrt;
					
					if(addr_type == CAT_BCC)
						goto cad_continue;
					
					addrt = (gchar *)node->data;

					/* Put a comma to separate multiples.
					 */
					if(addr_count)
					{
						fprintf(fp, ", ");
						line += 2;
						last_addr = 0;
					}
					
					/* Check the address text for a "<"
					 * character.  If it exists, and the
					 * following ">" exists, and the contained
					 * string is not 0-length, then just dump
					 * this string out as a valid address.
					 */
					if(c = strchr(addrt, '<'))
					{
						gchar *e;
						
						if(e = strchr(c, '>'))
						{
							if(e - c > 1)
							{
								gchar *addr_name, *addr_addr;
								
								/* Check if it will exceed nominal line length.
								 */
								if((line + strlen(addrt) + 3) > 70)
								{
									fprintf(fp, "%s    ", eoltypes[prefs.EOLType]);
									fflush(fp);
									line = 4;
								}
								
								*c = 0;
								addr_name = g_strdup(addrt);
								*c = '<';
								
								/* Output the name, if there's anything
								 * left.
								 */
								g_strstrip(addr_name);
								if(*addr_name)
								{
									work = g_strdup_printf("\"%s\"", addr_name);
									line += header_encode(fp, work);
									g_free(work);
								}
								g_free(addr_name);
								
								line += header_encode(fp, c);
							
								goto cad_continue;
							}
						}
					}
					
					/* There was not a "<" delimiter, so check
					 * for "@" now.  If found, we assume it's a
					 * valid, manually entered address.
					 */
					if(c = strchr(addrt, '@'))
					{
						/* Check if it will exceed nominal line length.
						 */
						if((line + strlen(addrt) + 3) > 70)
						{
							fprintf(fp, "%s    ", eoltypes[prefs.EOLType]);
							fflush(fp);
							line = 4;
						}
						
						line += header_encode(fp, addrt);
					
						goto cad_continue;
					}
					
					/* Dropped through, so we need to process
					 * this address a bit more.
					 */




#if 0 // FIXME
					/* Get a matching address or addresses.
					 */
					if(ae_list = address_match_name(addrt, FALSE))
					{
						/* Check if there is more than one match.
						 */
						if(g_list_length(ae_list) > 1)
						{
							/* Ask the user to select one.
							 * If CF_AUTO is set, then the first one
							 * will be returned, since we'll pass TRUE
							 * for the silent parameter.
							 */
							ae = address_select_one_from_many_matches(ae_list, comp->Window, comp->Flags & CF_AUTO);
						}
						else
							ae = (struct AddressEntry *)g_list_nth_data(ae_list->List, 0);
							
						if(ae)
						{
							gchar first_name[201], last_name[201];
							
							address_smart_get_first_last_name(ae, first_name, sizeof(first_name), last_name, sizeof(last_name));
							
							/* Should we start a new line?
							 */
							if((line + strlen(first_name) + strlen(last_name) + 3) > 70)
							{
								fprintf(fp, "%s    ", eoltypes[prefs.EOLType]);
								line = 4;
							}

							/* Check if this is a group.
							 */
							if(ae->Members)
							{
								struct AddressEntry *aem;
								
								/* Walk the list.
								 */
								node = g_list_first(ae->Members->List);
								while(node)
								{
									gchar *m_addr, m_first[201], m_last[201];
									
									aem = (struct AddressEntry *)node->data;
									
									/* Try to get the first and last name.
									 */
									address_smart_get_first_last_name(aem, m_first, sizeof(m_first), m_last, sizeof(m_last));
											
									/* Get the "email" field.
									 */
									if(m_addr = address_get_field(aem, "Email"))
									{
										/* If multiple addresses are available for
										 * this recipient, ask which one to use.
										 */
										if(strchr(m_addr, ','))
										{
											gchar work[201], *m_addr;
											
											g_snprintf(work, sizeof(work), "%s %s", m_first, m_last);
											
											/* Ask the user to select from multiple possible
											 * addresses for this entry.
											 */
											address_request_recipient(work, m_addr, addr, sizeof(addr) - 1, comp->Flags & CF_AUTO);
										}
										else
											strncpy(addr, m_addr, sizeof(addr) - 1);

#if 0	// FIXME
										if(!*addr)
										{    // request an address if one is missing
											GtkWidget *dialog;
											gchar prompt[201];

											g_snprintf(prompt, sizeof(prompt), _("Input an email address for:\n%s %s"), m_first, m_last);
											if(dialog = gnome_request_string_dialog_parented(prompt, addr, sizeof(addr), 0))
											{
											}
											else
												goto aem_continue;

											if(!*addr)
												goto aem_continue;
										}
#endif

										if(sub_addr_count)
										{
											fprintf(fp, ", ");
											line += 2;
										}

										if((line + strlen(m_first) + strlen(m_last) + 3) > 70)
										{
											work = g_strdup_printf("%s    ", eoltypes[prefs.EOLType]);
											fprintf(fp, work);
											g_free(work);
											line = 4;
										}

										work = g_strdup_printf("\"%s %s\"", m_first, m_last);
										line += header_encode(fp, work);
										g_free(work);

										if((line + strlen(m_addr) + 3) > 70)
										{
											work = g_strdup_printf("%s    ", eoltypes[prefs.EOLType]);
											fprintf(fp, work);
											g_free(work);
											line = 4;
										}

										work = g_strdup_printf(" <%s>", m_addr);
										line += header_encode(fp, work);
										g_free(work);

										sub_addr_count++;
									}
								
									aem_continue:
									node = g_list_next(node);
								}

								last_addr = 2;
							}
							else
							{	/* Not a group.
								 */
								work = g_strdup_printf("\"%s %s\"", ae->FirstName, ae->LastName);
								line += header_encode(fp, work);
								g_free(work);

								/* If multiple addresses are available for
								 * this recipient, ask which one to use.
								 */
								if(strchr(ae->Address, ','))
								{
									gchar work[201];
									
									g_snprintf(work, sizeof(work), "%s %s", ae->FirstName, ae->LastName);
									address_request_recipient(work, ae->Address, addr, sizeof(addr) - 1);
								}
								else
									strncpy(addr, ae->Address, sizeof(addr) - 1);

#if 0	// FIXME
								if(!*addr)
								{    // request an address if one is missing
									gchar prompt[201];

									g_snprintf(prompt, sizeof(prompt), , ae->FirstName, ae->LastName);
									if(!RequestString(prompt, addr, sizeof(addr), 0))
										goto cad_continue;

									if(!*addr)
										goto cad_continue;
								}
#endif

								if((line + strlen(addr) + 3) > 70)
								{
									work = g_strdup_printf("%s    ", eoltypes[prefs.EOLType]);
									fprintf(fp, work);
									g_free(work);
									line = 4;
								}

								work = g_strdup_printf(" <%s>", addr);
								line += header_encode(fp, work);
								g_free(work);
							}
						}
					}
					else
					{	/* No matching entry in the address book.
						 */
						 
						/* If multiple addresses are available for
						 * this recipient, ask which one to use.
						 */
#if 0
						if(strchr(addrt, ','))
						{
							address_request_recipient(caddr->Name, caddr->Address, addr, sizeof(addr) - 1, comp->Flags & CF_AUTO);
						}
						else
							strncpy(addr, caddr->Address, sizeof(addr) - 1);
#endif

#if 0	// FIXME
						if(!*addr)
						{    // request an address if one is missing
							gchar prompt[201];

							sprintf(prompt, LocaleString(ETXT_COMPOSE_INPUTADDRESS), caddr->Name, "");
							if(!RequestString(prompt, addr, sizeof(addr), 0))
								goto cad_continue;

							if(!*addr)
								goto cad_continue;
						}
#endif

						if((line + strlen(caddr->Name)) > 70)
						{
							fprintf(fp, "%s    ", eoltypes[prefs.EOLType]);
							line = 4;
						}

						if(strchr(caddr->Name, '@'))
						{
							work = g_strdup_printf("<%s>", caddr->Name);
							line += header_encode(fp, work);
							g_free(work);
						}
						else
						{
							if(*caddr->Name)
							{
								work = g_strdup_printf("\"%s\"", caddr->Name);
								line += header_encode(fp, work);
								g_free(work);
							}
						}

						if(!strchr(caddr->Name, '@'))
						{
							if((line + strlen(addr) + 3) > 70)
							{
								fprintf(fp, "%s    ", eoltypes[prefs.EOLType]);
								line = 4;
							}

							work = g_strdup_printf(" <%s>", addr);
							line += header_encode(fp, work);
							g_free(work);
						}
					}
#endif

					/* Resuming point.
					 */
					cad_continue:

					addr_count++;

					/* Write to "addr" file.
					 */
					switch(addr_type)
					{
						case CAT_TO:
							fprintf(addr_fp, "To: ");
							break;
						case CAT_CC:
							fprintf(addr_fp, "Cc: ");
							break;
						case CAT_BCC:
							fprintf(addr_fp, "Bcc: ");
							break;
					}

					fprintf(addr_fp, "%s\n", addrt);
				}

				fprintf(fp, eoltypes[prefs.EOLType]);
			}

			if(*comp->t.ReplyTo)
			{
				work = g_strdup_printf("Reply-To: %s%s", comp->t.ReplyTo, eoltypes[prefs.EOLType]);
				header_encode(fp, work);
				g_free(work);
			}

			if(comp->Flags & CF_RETURNRECEIPT)
			{
				work = g_strdup_printf("Return-Receipt-To: %s%s", uprefs.Address, eoltypes[prefs.EOLType]);
				header_encode(fp, work);
				g_free(work);
			}

			// date
			{
				gchar *weekdays[] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat",};
				gchar *months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec",};

				fprintf(fp, "Date: %s, %d %s %d %2.2d:%2.2d:%2.2d%s",
							weekdays[tm->tm_wday],
							tm->tm_mday, months[tm->tm_mon], tm->tm_year,
							tm->tm_hour, tm->tm_min, tm->tm_sec,
							eoltypes[prefs.EOLType]);
			}

			text = gtk_editable_get_chars(GTK_EDITABLE(comp->w.Subject), 0, -1);
			work = g_strdup_printf("Subject: %s%s", *text ? text : "<No subject>", eoltypes[prefs.EOLType]);
			header_encode(fp, work);
			g_free(work);
			g_free(text);

			fprintf(fp, "MIME-Version: 1.0%s", eoltypes[prefs.EOLType]);
			fprintf(fp, "X-Mailer: Eucalyptus for Linux " VERSION " " __DATE__ "%s",
					eoltypes[prefs.EOLType]);

			{
				gchar *domain;

				if(domain = strchr(uprefs.Address, '@'))
					domain++;
				else
					domain = "eucalyptus";

				fprintf(fp, "Message-ID: <%4.4d%2.2d%2.2d%2.2d%2.2d%2.2d.%lX@%s>%s",
						tm->tm_year, tm->tm_mon, tm->tm_mday,
						tm->tm_hour, tm->tm_min, tm->tm_sec,
						(long)getpid(),
						domain,
						eoltypes[prefs.EOLType]);
			}

			// reference info
			if(*comp->t.Reference)
			{
				if(comp->Flags & CF_REPLIEDMSG)
				{
					fprintf(fp, "In-Reply-To: %s%s", comp->t.Reference, eoltypes[prefs.EOLType]);
				}
				else
				{
					fprintf(fp, "References: %s%s", comp->t.Reference, eoltypes[prefs.EOLType]);
				}
			}

			fprintf(fp, "Priority: %s%s", priorities[comp->MsgPriority], eoltypes[prefs.EOLType]);

			// custom headers
			node = g_list_first(comp->CustomHeaders.List);
			while(node)
			{
				gchar *str = (gchar *)node->data;
				
				if(str && *str)
				{
					gchar *c, *x;

					if(c = strchr(str, ':'))
					{
						*c = '\0';
						x = strchr(str, ' ');
						*c = ':';

						if(!x)
						{
							work = g_strdup_printf("%s%s%s",
									strncasecmp(str, "x-", 2) ? "X-" : "",
									str,
									eoltypes[prefs.EOLType]);
							header_encode(fp, work);
							g_free(work);
						}
						else
						{
							success = FALSE;
							goto fp_out;
						}
					}
				}
				
				node = g_list_next(node);
			}

			if(comp->Flags & (CF_SIGN | CF_ENCRYPT))
			{
				fprintf(fp, "Content-Type: multipart/%s; boundary=\"%s\";%s"
						  "              micalg=pgp-md5; protocol=\"application/pgp-%s\"%s",
						  (comp->Flags & CF_ENCRYPT) ? "encrypted" : "signed",
						  comp->Boundary,
						  eoltypes[prefs.EOLType],
						  (comp->Flags & CF_ENCRYPT) ? "encrypted" : "signature",
						  eoltypes[prefs.EOLType]);
				multipart = TRUE;
			}
			else
			{
				if(g_list_length(comp->Attachments.List))
				{    // attachments
					fprintf(fp, "Content-Type: multipart/mixed; boundary=\"%s\"%s",
							  comp->Boundary,
							  eoltypes[prefs.EOLType]);
					multipart = TRUE;
				}
			}

			if(multipart)
			{
				fprintf(fp, eoltypes[prefs.EOLType]);
				fprintf(fp, "This is a MIME multipart message. If you see this text, you are not%s"
						  "using MIME aware mail software. To decode the attachments in this message,%s"
						  "you may use the \'metamail\' package.%s"
						  "%s"
						  "This message was written with Eucalyptus.%s",
						  eoltypes[prefs.EOLType], eoltypes[prefs.EOLType], eoltypes[prefs.EOLType], eoltypes[prefs.EOLType], eoltypes[prefs.EOLType]);
				fprintf(fp, "%s--%s%s",
						eoltypes[prefs.EOLType],
						comp->Boundary,
						eoltypes[prefs.EOLType]);
			}

			if(comp->Flags & (CF_SIGN | CF_ENCRYPT))
			{
#if 0	// FIXME
				FILE *tmp_fp;
				gchar fname[51], pgpfile[51], cmdline[201], switches[21];

				GetTempFilename("T:", fname, sizeof(fname));

				if(tmp_fp = fopen(fname, MODE_WRITE, ASYNCIO_BUF, SysBase, DOSBase))
				{
					FILE *pgp_fp;

					if(IsListEmpty(&comp->Attachments))
					{    /* No attachments, so just write out main text,
						 * call PGP to sign/encrypt, and merge data
						 * as required
						 */
						error = compose_write_contents(tmp_fp, comp);
					}
					else
					{
						gchar boundary[31];

						srand((time(NULL) + 2) * time(NULL));
						memset(boundary, 0, sizeof(boundary));
						for(x=0; x<sizeof(boundary)-1; x++)
							boundary[x] = charset[(rand() + 20) % strlen(charset)];

						fprintf(tmp_fp, "Content-Type: multipart/mixed; boundary=\"%s\"%s",
								  boundary,
								  eoltypes[prefs.EOLType]);
						fprintf(tmp_fp, "%s--%s%s",
								eoltypes[prefs.EOLType],
								boundary,
								eoltypes[prefs.EOLType]);

						if(!(error = compose_write_msg_contents(comp, tmp_fp)))
						{
							if(!IsListEmpty(&comp->Attachments))
								error = compose_write_msg_attachments(tmp_fp, msgfile, comp, boundary);

							fprintf(tmp_fp, "%s--%s--%s",
									eoltypes[prefs.EOLType],
									boundary,
									eoltypes[prefs.EOLType]);
						}
					}

					fclose(tmp_fp);

					if(!error)
					{
						GetTempFilename("T:", pgpfile, sizeof(pgpfile));

						strncpy(switches, "-", sizeof(switches) - 1);
						if(comp->Flags & CF_SIGN)
							strncat(switches, "s", sizeof(switches) - 1);
						if(comp->Flags & CF_ENCRYPT)
							strncat(switches, "e", sizeof(switches) - 1);
						else
							strncat(switches, "b", sizeof(switches) - 1);
						strncat(switches, "a", sizeof(switches) - 1);

						if(comp->Flags & (CF_ENCRYPT | CF_ENCRYPTTOSELF))
							strncat(switches, " +encrypttoself", sizeof(switches) - 1);

						if(oprefs.UsePGP5)
						{
							if(comp->Flags & CF_ENCRYPT)
								strncpy(cmdline, "PGPe", sizeof(cmdline) - 1);
							else
								strncpy(cmdline, "PGPs", sizeof(cmdline) - 1);
						}
						else
							strncpy(cmdline, oprefs.PGPPath, sizeof(cmdline) - 1);
						strncat(cmdline, " ", sizeof(cmdline) - 1);
						strncat(cmdline, switches, sizeof(cmdline) - 1);
						if(oprefs.UsePGP5)
							strncat(cmdline, " +verbose=0 ", sizeof(cmdline) - 1);
						else
							strncat(cmdline, " -z ", sizeof(cmdline) - 1);
						strncat(cmdline, fname, sizeof(cmdline) - 1);
						if(comp->Flags & CF_ENCRYPT)
						{    // add user IDs
							gchar *text;

							for(caddr = (CompAddress *)comp->Recipients.lh_Head;
							    caddr->Node.ln_Succ;
							    caddr = (CompAddress *)caddr->Node.ln_Succ)
							{
								struct AddressEntry *ae;
								gchar prompt[201], keyid[81];

								if(ae = MatchAddressName(caddr->Name, FALSE))
								{
									if(strlen(ae->KeyID))
									{
										strncat(cmdline, " \"", sizeof(cmdline) - 1);
										strncat(cmdline, ae->KeyID, sizeof(cmdline) - 1);
										strncat(cmdline, "\"", sizeof(cmdline) - 1);
										continue;
									}
								}

								// ask for a keyID
								sprintf(prompt, LocaleString(ETXT_COMPOSE_INPUTKEYID), caddr->Name);
								memset(keyid, 0, sizeof(keyid));
								strncpy(keyid, caddr->Name, sizeof(keyid) - 1);
								if(!RequestString(prompt, keyid, sizeof(keyid), 0))
									continue;

								strncat(cmdline, " \"", sizeof(cmdline) - 1);
								strncat(cmdline, keyid, sizeof(cmdline) - 1);
								strncat(cmdline, "\"", sizeof(cmdline) - 1);
							}

							GetAttr(STRINGA_TextVal, comp->EncryptTo, (ULONG *)&text);
							if(text)
							{
								strncat(cmdline, " ", sizeof(cmdline) - 1);
								strncat(cmdline, text, sizeof(cmdline) - 1);
							}
						}
						strncat(cmdline, " -o ", sizeof(cmdline) - 1);
						strncat(cmdline, pgpfile, sizeof(cmdline) - 1);
						if(comp->Flags & CF_ENCRYPT)
						{
							gchar *t;

							GetAttr(STRINGA_TextVal, comp->EncryptTo, (ULONG *)&t);
							strncat(cmdline, " ", sizeof(cmdline) - 1);
							if(t)
								strncat(cmdline, t, sizeof(cmdline) - 1);
						}

						if(pgp_fp = fopen("CON:0/0/640/200/Eucalyptus Output/AUTO/SIMPLE", MODE_OLDFILE))
						{
							glong e;
							gchar pgptries = 0;

							switch(e = SystemTags(cmdline,
									SYS_Input,		pgp_fp,
									SYS_Output,    	NULL,
									TAG_END))
							{
								case -1:
									error = err_UNABLETOEXECUTECMD;
									break;
								case 0:
									// read msg file
									// Flush(fp);
									if(comp->Flags & CF_ENCRYPT)
									{
										work = g_strdup_printf("Content-Type: application/pgp-encrypted%s"
												"%s"
												"Version: 1%s",
												eoltypes[prefs.EOLType], eoltypes[prefs.EOLType], eoltypes[prefs.EOLType]);
										fprintf(fp, work);
										g_free(work);
									}
									else
									{
										if(tmp_fp = fopen(fname, MODE_READ, ASYNCIO_BUF, SysBase, DOSBase))
										{
											gchar in[201];

											while(FGetsAsync(tmp_fp, in, sizeof(in) - 1))
												fprintf(fp, in);

											fclose(tmp_fp);
										}
									}

									// Flush(fp);
									work = g_strdup_printf("%s--%s%s",
											eoltypes[prefs.EOLType],
											comp->Boundary,
											eoltypes[prefs.EOLType]);
									fprintf(fp, work);
									g_free(work);
									work = g_strdup_printf("Content-Type: application/%s%s",
											  (comp->Flags & CF_ENCRYPT) ? "octet-stream" : "pgp-signature",
											  eoltypes[prefs.EOLType]);
									fprintf(fp, work);
									g_free(work);
									work = g_strdup_printf("%s", eoltypes[prefs.EOLType]);
									fprintf(fp, work);
									g_free(work);

									// read PGP-generated file
									open_pgpfile:
									if(tmp_fp = fopen(pgpfile, MODE_READ, ASYNCIO_BUF, SysBase, DOSBase))
									{
										gchar in[201];

										while(fgets(tmp_fp, in, sizeof(in) - 1))
											fprintf(fp, in);

										fclose(tmp_fp);
									}
									else
									{
										if(!pgptries)
										{
											strncat(pgpfile, ".asc", sizeof(pgpfile) - 1);
											pgptries++;
											goto open_pgpfile;
										}

										success = FALSE;
									}
									break;
									
								default:
									success = FALSE;
									break;
							}

							fclose(pgp_fp);
						}
						else
						{
							success = FALSE;
						}
					}
				}
				else
				{
					success = FALSE;
				}
#endif
			}
			else
			{
				if(compose_write_contents(comp, fp))
				{
					if(g_list_length(comp->Attachments.List))
					{
						if(!compose_write_attachments(comp, fp, msgfile, comp->Boundary, infofile))
						{
							   success = FALSE;
						}
					}
				}
			}

			if(multipart)
			{
				fprintf(fp, "%s--%s--%s%s",
						eoltypes[prefs.EOLType],
						comp->Boundary,
						eoltypes[prefs.EOLType], eoltypes[prefs.EOLType]);
				fprintf(fp, "For information about Eucalyptus, visit%s"
						  "<http://www.isengard-dev.org/>%s",
						  eoltypes[prefs.EOLType], eoltypes[prefs.EOLType]);
			}

			fp_out:
			fclose(addr_fp);
		}
		else
		{
			if(!(comp->Flags & CF_AUTO))
				euc_request(comp->Window,
						_("Eucalyptus error"),
						_("Unable to create message."),
						euc_ok_button, 0,
						0);
			success = FALSE;
		}

		fclose(fp);
	}
	else
	{
		if(!(comp->Flags & CF_AUTO))
			euc_request(comp->Window,
					_("Eucalyptus error"),
					_("Unable to create message."),
					euc_ok_button, 0,
					0);
		success = FALSE;
	}

	folder->Data.Flags |= (FFF_NEWMSGS | FFF_INDEXMODIFIED | FFF_UPDATEMAIN);
	
	/* Clean up.
	 */
	g_free(dir);
	g_free(msgfile);
	g_free(addrfile);
	g_free(infofile);

	return success;
}

gboolean compose_write_contents(Composition *comp, FILE *fp)
{
	glong size;
	gchar *eoltypes[] = {"\r\n", "\r", "\n", };
	gchar *cte_type[] = {"8bit", "quoted-printable",};
	gchar *cs_type[] = {"ISO-8859-1", "us-ascii",};
	gchar *text;
	gboolean success = TRUE;

	fprintf(fp, "Content-Type: text/plain; charset=%s%s", (prefs.UseQuotedPrintable ? cs_type[1] : cs_type[0]), eoltypes[prefs.EOLType]);
	fprintf(fp, "Content-Transfer-Encoding: %s%s", (prefs.UseQuotedPrintable ? cte_type[1] : cte_type[0]), eoltypes[prefs.EOLType]);
	fprintf(fp, eoltypes[prefs.EOLType]);

	/* Do body "header" file.
	 * NOT IMPLEMENTED
	 */

	/* Do text.
	 */
	text = gtk_editable_get_chars(GTK_EDITABLE(comp->w.Textfield), 0, -1);
	header_write_quoted_printable(fp, text, strlen(text), 0);
	g_free(text);

	/* Paranoia.
	 */
	fflush(fp);

	/* Do body "footer".
	 * NOT IMPLEMENTED
	 */

	/* Do tagline.
	 */
	if(comp->Flags & CF_USETAGLINE)
	{
		text = gtk_editable_get_chars(GTK_EDITABLE(comp->w.Tagline), 0, -1);
		fprintf(fp, "%s%s... ", eoltypes[prefs.EOLType], eoltypes[prefs.EOLType]);
		header_write_quoted_printable(fp, text, strlen(text), 4);
		fprintf(fp, eoltypes[prefs.EOLType]);
	}

	/* Do sig file.
	 */
	if(comp->Flags & CF_USESIGFILE)
	{
		text = gnome_file_entry_get_full_path(GNOME_FILE_ENTRY(comp->w.SignatureFile), TRUE);
		if(text && *text)
		{
			gchar in[201];
			FILE *sig_fp;

			if(sig_fp = fopen(text, "r"))
			{
				fprintf(fp, "%s--%s%s",
						eoltypes[prefs.EOLType],
						prefs.UseQuotedPrintable ? "=20" : " ",
						eoltypes[prefs.EOLType]);

				while(fgets(in, sizeof(in) - 1, sig_fp))
				{
					header_write_quoted_printable(fp, in, strlen(in), 0);
				}

				fclose(sig_fp);
			}
		}
	}

	return success;
}

gboolean compose_write_attachments(Composition *comp, FILE *fp, gchar *msgfile, gchar *boundary, gchar *infofile)
{
	CompAttachment *catt;
	GList *node;
	FILE *att_fp;
	gchar *eoltypes[] = {"\r\n", "\r", "\n", };
	gchar *cte_type[] = {"8bit", "quoted-printable",};
	gboolean success = TRUE;
	FILE *info_fp;

	info_fp = fopen(infofile, "w");

	   for(node = g_list_first(comp->Attachments.List); node; node = g_list_next(node))
	   {
			 catt = (CompAttachment *)node->data;

			 fprintf(fp, "%s--%s%s",
					eoltypes[prefs.EOLType],
					boundary,
					eoltypes[prefs.EOLType]);

			 if((catt->Type & CONTENTTYPE_MASK) == CONTENTTYPE_TEXT)
			 {
				gchar *text;

				switch(catt->Type & CONTENTSUBTYPE_MASK)
				{
					case CONTENTSUBTYPE_TEXT_PLAIN:
					default:
						fprintf(fp, "Content-Type: text/plain; charset=ISO-8859-1; name=\"%s\"%s", catt->Name, eoltypes[prefs.EOLType]);
						break;
					case CONTENTSUBTYPE_TEXT_ENRICHED:
						fprintf(fp, "Content-Type: text/enriched; charset=ISO-8859-1; name=\"%s\"%s", catt->Name, eoltypes[prefs.EOLType]);
						break;
					case CONTENTSUBTYPE_TEXT_HTML:
						fprintf(fp, "Content-Type: text/html; charset=ISO-8859-1; name=\"%s\"%s", catt->Name, eoltypes[prefs.EOLType]);
						break;
				}
				fprintf(fp, "Content-Transfer-Encoding: %s%s", (prefs.UseQuotedPrintable ? cte_type[1] : cte_type[0]), eoltypes[prefs.EOLType]);
				fprintf(fp, "Content-Disposition: attachment; filename=\"%s\"%s", catt->Name, eoltypes[prefs.EOLType]);
				if(*catt->Description)
				{
					fprintf(fp, "Content-Description: %s%s", catt->Description, eoltypes[prefs.EOLType]);
				}
				fprintf(fp, eoltypes[prefs.EOLType]);

				if(att_fp = fopen(catt->Filename, "r"))
				{
					gchar text[WORKLEN];

					if(info_fp)
						fprintf(info_fp, "%s\n", catt->Filename);

					while(fgets(text, sizeof(text) - 1, att_fp))
						header_write_quoted_printable(fp, text, strlen(text), 0);

					fclose(att_fp);
				}
				else
				{
					success = FALSE;
					break;
				}

				continue;
			 }
			 if((catt->Type & CONTENTTYPE_MASK) == CONTENTTYPE_MESSAGE)
			 {
				gchar *text;

				switch(catt->Type & CONTENTSUBTYPE_MASK)
				{
					case CONTENTSUBTYPE_MESSAGE_RFC822:
					default:
						fprintf(fp, "Content-Type: message/rfc822; name=\"%s\"%s", catt->Name, eoltypes[prefs.EOLType]);
						break;
					case CONTENTSUBTYPE_MESSAGE_PARTIAL:
						fprintf(fp, "Content-Type: message/partial; name=\"%s\"%s", catt->Name, eoltypes[prefs.EOLType]);
						break;
				}
				fprintf(fp, "Content-Transfer-Encoding: 7bit%s", eoltypes[prefs.EOLType]);
				fprintf(fp, "Content-Disposition: attachment; filename=\"%s\"%s", catt->Name, eoltypes[prefs.EOLType]);
				if(*catt->Description)
				{
					fprintf(fp, "Content-Description: %s%s", catt->Description, eoltypes[prefs.EOLType]);
				}
				fprintf(fp, eoltypes[prefs.EOLType]);

				if(att_fp = fopen(catt->Filename, "r"))
				{
					gchar text[WORKLEN];
					
					if(info_fp)
						fprintf(info_fp, "%s\n", catt->Filename);

					while(fgets(text, sizeof(text) - 1, att_fp))
						fprintf(fp, text);

					fclose(att_fp);
				}
				else
				{
					success = FALSE;
					break;
				}

				continue;
			 }
			 fprintf(fp, "Content-Type: %s; name=\"%s\"%s", *catt->ContentType ? catt->ContentType : "application/octet-stream", catt->Name, eoltypes[prefs.EOLType]);
			 fprintf(fp, "Content-Transfer-Encoding: base64%s", eoltypes[prefs.EOLType]);
			 fprintf(fp, "Content-Disposition: attachment; filename=\"%s\"%s", catt->Name, eoltypes[prefs.EOLType]);
			 if(*catt->Description)
			 {
				fprintf(fp, "Content-Description: %s%s", catt->Description, eoltypes[prefs.EOLType]);
			 }
			 fprintf(fp, eoltypes[prefs.EOLType]);

			 if(att_fp = fopen(catt->Filename, "r"))
			 {
				glong c, qpos = 0, triplet = 0, lpos = 0;
				glong mask_table[] = {0xFC0000, 0x3F000, 0xFC0, 0x3F,};
				gchar opos = 0;

				if(info_fp)
					fprintf(info_fp, "%s\n", catt->Filename);

				fflush(fp);

				while(-1 != (c = fgetc(att_fp)))
				{
					triplet |= (c << ((2 - qpos) * 8));
					qpos++;

					if(qpos == 3)
					{
						opos = 0;

						while(opos < 4)
						{
							fputc(base64set[(triplet & mask_table[opos]) >> ((3 - opos) * 6)], fp);
							lpos++;
							opos++;
						}

						qpos = 0;
						triplet = 0;
					}

					if(lpos > 60)
					{
						fprintf(fp, eoltypes[prefs.EOLType]);
						lpos = 0;
					}
				}

				if(qpos)
				{
					opos = 0;

					while(opos < (4 - (3 - qpos)))
					{
						fputc(base64set[(triplet & mask_table[opos]) >> ((3 - opos) * 6)], fp);
						opos++;
					}
					for(; qpos<3; qpos++)
						fprintf(fp, "=");
				}

				fprintf(fp, eoltypes[prefs.EOLType]);

				fclose(att_fp);
			 }
			 else
			 {
				success = FALSE;
				break;
			 }
	   }

	   if(info_fp)
		   fclose(info_fp);

	return success;
}

gboolean compose_postpone_msg(Composition *comp)
{
	gboolean success = FALSE;
	Folder *folder;

	if(folder = folder_get_special(FTF_POSTPONED))
	{
		gchar fname[FILENAME_MAX], *ftmp;
		FILE *fp;

		if(*comp->t.PostponedFile)
		{
			ftmp = folder_get_file_path(folder, comp->t.PostponedFile);
			strncpy(fname, ftmp, sizeof(fname) - 1);
			g_free(ftmp);
		}
		else
			folder_get_next_msg_num(folder, NULL, fname, sizeof(fname));

		if(fp = fopen(fname, "w"))
		{
			CompAttachment *catt;
			GList *node;
			gchar *text;
			gboolean use;

			text = gtk_editable_get_chars(GTK_EDITABLE(comp->w.AddrTo), 0, -1);
			fprintf(fp, "To: %s\n", text);
			g_free(text);

			text = gtk_editable_get_chars(GTK_EDITABLE(comp->w.AddrCc), 0, -1);
			fprintf(fp, "Cc: %s\n", text);
			g_free(text);

			text = gtk_editable_get_chars(GTK_EDITABLE(comp->w.AddrBcc), 0, -1);
			fprintf(fp, "Bcc: %s\n", text);
			g_free(text);

			text = gtk_editable_get_chars(GTK_EDITABLE(comp->w.ReplyTo), 0, -1);
			if(*text)
				fprintf(fp, "Reply-To: %s\n", text);
			g_free(text);

			text = gtk_editable_get_chars(GTK_EDITABLE(comp->w.Subject), 0, -1);
			fprintf(fp, "Subject: %s\n", text);
			g_free(text);

			for(node = g_list_first(comp->Attachments.List); node; node = g_list_next(node))
			{
				gchar encoding_table[] = {"Base64QP", };

				catt = (CompAttachment *)node->data;

				fprintf(fp, "Attachment: \"%s\"; \"%s\"; \"%s\"; \"%s\"; \"%s\"\n",
						catt->Name,
						catt->Filename,
						encoding_table[catt->Encoding],
						catt->ContentType,
						catt->Description);
			}

			text = gtk_editable_get_chars(GTK_EDITABLE(comp->w.Tagline), 0, -1);
			use = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(comp->w.UseTagline));
			if(use && *text)
			{
				fprintf(fp, "Tagline: %s\n", text);
			}
			g_free(text);

			text = gtk_editable_get_chars(GTK_EDITABLE(comp->w.SignatureFile), 0, -1);
			use = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(comp->w.UseSigFile));
			if(use && *text)
			{
				fprintf(fp, "Signature-File: %s\n", text);
			}
			g_free(text);

			fprintf(fp, "Sign: %s\n",
					  (comp->Flags & CF_SIGN) ? "Yes" : "No");

			fprintf(fp, "Encrypt: %s\n",
					  (comp->Flags & CF_ENCRYPT) ? "Yes" : "No");
			text = gtk_editable_get_chars(GTK_EDITABLE(comp->w.EncryptTo), 0, -1);
			if(*text)
			{
				fprintf(fp, "Encrypt-To: %s\n", text);
			}
			g_free(text);

			fprintf(fp, "\n");
			fflush(fp);

			text = gtk_editable_get_chars(GTK_EDITABLE(comp->w.Textfield), 0, -1);

			if(text)
				fprintf(fp, text);

			folder->Data.Flags |= (FFF_INDEXMODIFIED | FFF_UPDATEMAIN);

			fclose(fp);
		}
		else
			success = FALSE;
	}
	else
		success = FALSE;

	return success;
}

void compose_delete_postponed_msg(Composition *comp)
{
	Folder *folder;

	if(folder = folder_get_special(FTF_POSTPONED))
	{
//			DeleteFile(comp->t.PostponedFile);
	}
}

gboolean compose_resume_postponed_msg(MailMsg *msg)
{
	gboolean success = FALSE;
#if 0
	Composition *comp;
	CompAttachment *catt;
	LONG error = 0;
	UBYTE *text;
	BPTR dir, fp;

	if(!msg ||
	   !msg->Folder)
		return(0);

	if(dir = CurrentDir(msg->Folder->DirLock))
	{
		if(comp = (Composition *)AllocVec(sizeof(Composition), MEMF_CLEAR))
		{
			InitComposition(comp);

			if(text = (UBYTE *)AllocVec(WORKLEN * 2, MEMF_CLEAR))
			{
				if(fp = Open(msg->Index.Filename, MODE_OLDFILE))
				{
					strncpy(comp->t.PostponedFile, msg->Index.Filename, sizeof(comp->t.PostponedFile) - 1);

					while(FGets(fp, text, (WORKLEN * 2) - 1))
					{
						UBYTE *c, *d;
						BYTE type;

						if(c = strchr(text, '\n'))
							*c = '\0';
						if(c = strchr(text, '\r'))
							*c = '\0';

						if(!strlen(text))
							break;

						if(c = strchr(text, ':'))
						{
							*c = '\0';
							d = stpblk(c + 1);
						}
						else
							continue;

						type = -1;
						if(!strcmp(text, "To"))
							type = CAT_TO;
						else if(!strcmp(text, "Cc"))
							type = CAT_CC;
						else if(!strcmp(text, "Bcc"))
							type = CAT_BCC;

						if(type != -1)
						{
							AddCompAddress(comp, d, strchr(d, '<') ? ACAFT_PARSE : ACAFT_NAME, type);
							continue;
						}

						if(!strcmp(text, "Attachment"))
						{
							if(catt = (CompAttachment *)AllocListBrowserNode(2,
									LBNA_NodeSize, 		sizeof(CompAttachment),
									LBNA_Column,   		0,
										LBNCA_Text,    		"",
										LBNCA_Justification,	LCJ_LEFT,
									LBNA_Column,   		1,
										LBNCA_Text,    		"",
										LBNCA_Justification,	LCJ_LEFT,
									TAG_END))
							{
								if(c = strchr(d, '\"'))
								{    // name
									d = c + 1;
									if(c = strchr(d, '\"'))
									{
										*c = '\0';
										strncpy(catt->Name, d, sizeof(catt->Name) - 1);
										d = c + 1;

										if(c = strchr(d, '\"'))
										{    // fname
											d = c + 1;
											if(c = strchr(d, '\"'))
											{
												*c = '\0';
												strncpy(catt->Filename, d, sizeof(catt->Filename) - 1);
												d = c + 1;

												if(c = strchr(d, '\"'))
												{    // encoding
													d = c + 1;

													if(c = strchr(d, '\"'))
													{
														*c = '\0';
														if(!strcmp(d, "Base64QP"))
															catt->Encoding = CAE_BASE64QP;
														d = c + 1;

														if(c = strchr(d, '\"'))
														{    // content-type
															d = c + 1;

															if(c = strchr(d, '\"'))
															{
																LONG i;

																*c = '\0';
																strncpy(catt->ContentType, d, sizeof(catt->ContentType) - 1);

																for(i=0; ContentTable[i].Name; i++)
																{
																	if(!stricmp(catt->ContentType, ContentTable[i].Name))
																	{
																		catt->Type = ContentTable[i].Value;
																		break;
																	}
																}

																d = c + 1;

																if(c = strchr(d, '\"'))
																{    // desc
																	d = c + 1;

																	if(c = strchr(d, '\"'))
																	{
																		*c = '\0';
																		strncpy(catt->Description, d, sizeof(catt->Description) - 1);
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}

								SetListBrowserNodeAttrs(catt,
										LBNA_Column,   	0,
											LBNCA_Text,    	catt->Name,
										LBNA_Column,   	0,
											LBNCA_Text,    	catt->ContentType,
										TAG_END);

								AddTail(&comp->Attachments, (struct Node *)catt);
							}
							continue;
						}
						if(!strcmp(text, "Subject"))
						{
							strncpy(comp->Subject, d, sizeof(comp->Subject) - 1);
							continue;
						}
						if(!strcmp(text, "Tagline"))
						{
							strncpy(comp->TaglineText, d, sizeof(comp->TaglineText) - 1);
							continue;
						}
						if(!strcmp(text, "Signature-File"))
						{
							strncpy(comp->SignatureFile, d, sizeof(comp->SignatureFile) - 1);
							continue;
						}
						if(!strcmp(text, "Sign"))
						{
							if(!strcmp(d, "Yes"))
								comp->Flags |= CF_SIGN;
							else
								comp->Flags &= ~CF_SIGN;
							continue;
						}
						if(!strcmp(text, "Encrypt"))
						{
							if(!strcmp(d, "Yes"))
								comp->Flags |= CF_ENCRYPT;
							else
								comp->Flags &= ~CF_ENCRYPT;
							continue;
						}
						if(!strcmp(text, "Encrypt-To"))
						{
							strncpy(comp->EncryptToText, d, sizeof(comp->EncryptToText) - 1);
							continue;
						}
					}

					Close(fp);
				}
				else
					error = err_UNABLETOOPENFILE;

				FreeVec(text);
			}

			if(error)
				FreeComposition(comp);
			else
			{
				comp->Flags |= CF_RESUMEDMSG;
				AddCompose(comp);
			}
		}

		CurrentDir(dir);
	}
#endif

	return success;
}
