/*
 * Eucalyptus
 * by Paul A. Schifferer
 *
 * Module: content.c
 *   	 Message content functions
 *
 * Copyright  1997-9 Isengard Developments
 * All rights reserved.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>

#include "eucalyptus.h"

/* Content-Type functions.
 */
glong content_type_verify(const gchar *hdr)
{
     struct ContentType content_table[] =
     {
          {"text/plain",                CONTENTTYPE_TEXT,},
          {"text/enriched",             CONTENTTYPE_TEXT | CONTENTSUBTYPE_TEXT_ENRICHED,},
          {"text/html",                 CONTENTTYPE_TEXT | CONTENTSUBTYPE_TEXT_HTML,},
          {"text/",                     CONTENTTYPE_TEXT,},
          {"image/jpeg",                CONTENTTYPE_IMAGE | CONTENTSUBTYPE_IMAGE_JPEG,},
          {"image/gif",                 CONTENTTYPE_IMAGE | CONTENTSUBTYPE_IMAGE_GIF,},
          {"image/png",                 CONTENTTYPE_IMAGE | CONTENTSUBTYPE_IMAGE_PNG,},
          {"image/x-png",               CONTENTTYPE_IMAGE | CONTENTSUBTYPE_IMAGE_PNG,},
          {"image/x-iff",               CONTENTTYPE_IMAGE | CONTENTSUBTYPE_IMAGE_XIFF,},
          {"image/",                    CONTENTTYPE_IMAGE,},
          {"audio/basic",               CONTENTTYPE_AUDIO | CONTENTSUBTYPE_AUDIO_BASIC,},
          {"audio/x-8svx",              CONTENTTYPE_AUDIO | CONTENTSUBTYPE_AUDIO_X8SVX,},
          {"audio/x-wav",               CONTENTTYPE_AUDIO | CONTENTSUBTYPE_AUDIO_XWAVE,},
          {"audio/",                    CONTENTTYPE_AUDIO,},
          {"video/mpeg",                CONTENTTYPE_VIDEO | CONTENTSUBTYPE_VIDEO_MPEG,},
          {"video/x-avi",               CONTENTTYPE_VIDEO | CONTENTSUBTYPE_VIDEO_XAVI,},
          {"video/x-quicktime",         CONTENTTYPE_VIDEO | CONTENTSUBTYPE_VIDEO_XQUICKTIME,},
          {"video/",                    CONTENTTYPE_VIDEO,},
          {"application/octet-stream",  CONTENTTYPE_APPLICATION | CONTENTSUBTYPE_APPLICATION_OCTETSTREAM,},
          {"application/pgp-encrypted", CONTENTTYPE_APPLICATION | CONTENTSUBTYPE_APPLICATION_PGPENCRYPTED,},
          {"application/pgp-signature", CONTENTTYPE_APPLICATION | CONTENTSUBTYPE_APPLICATION_PGPSIGNATURE,},
          {"application/pgp-keys",      CONTENTTYPE_APPLICATION | CONTENTSUBTYPE_APPLICATION_PGPKEYS,},
          {"application/postscript",    CONTENTTYPE_APPLICATION | CONTENTSUBTYPE_APPLICATION_POSTSCRIPT,},
          {"application/",              CONTENTTYPE_APPLICATION,},
          {"multipart/mixed",           CONTENTTYPE_MULTIPART | CONTENTSUBTYPE_MULTIPART_MIXED,},
          {"multipart/report",          CONTENTTYPE_MULTIPART | CONTENTSUBTYPE_MULTIPART_REPORT,},
          {"multipart/alternative",     CONTENTTYPE_MULTIPART | CONTENTSUBTYPE_MULTIPART_ALTERNATIVE,},
          {"multipart/parallel",        CONTENTTYPE_MULTIPART | CONTENTSUBTYPE_MULTIPART_PARALLEL,},
          {"multipart/signed",          CONTENTTYPE_MULTIPART | CONTENTSUBTYPE_MULTIPART_SIGNED,},
          {"multipart/encrypted",       CONTENTTYPE_MULTIPART | CONTENTSUBTYPE_MULTIPART_ENCRYPTED,},
          {"multipart/transport",       CONTENTTYPE_MULTIPART | CONTENTSUBTYPE_MULTIPART_TRANSPORT,},
          {"multipart/",                CONTENTTYPE_MULTIPART,},
          {"message/partial",           CONTENTTYPE_MESSAGE | CONTENTSUBTYPE_MESSAGE_PARTIAL,},
          {"message/external-body",     CONTENTTYPE_MESSAGE | CONTENTSUBTYPE_MESSAGE_EXTERNALBODY,},
          {"message/rfc822",            CONTENTTYPE_MESSAGE | CONTENTSUBTYPE_MESSAGE_RFC822,},
          {"message/delivery-status",   CONTENTTYPE_MESSAGE | CONTENTSUBTYPE_MESSAGE_DELIVERYSTATUS,},
          {"message/",                  CONTENTTYPE_MESSAGE,},
          {"text",                      CONTENTTYPE_TEXT,},
          {NULL,},
     };
	glong i, value = -1;

	if(!hdr)
		return -1;

	for(i=0; content_table[i].Name; i++)
	{
		gchar *ctype;

		if(ctype = content_type_find_param(hdr, content_table[i].Name, TRUE))
			value = content_table[i].Value;

		g_free(ctype);

		if(value != -1)
			break;
	}

	if(value == -1)
     	value = (CONTENTTYPE_APPLICATION | CONTENTSUBTYPE_APPLICATION_OCTETSTREAM);

     return value;
}

gchar *content_type_find_param(const gchar *hdr, const gchar *srch, gboolean partial)
{
	gchar *text, *param = NULL, *p, *e;
	GList *node;

	text = g_strdup(hdr);

	p = text;
	for(;;)
	{
		if(e = strchr(p, ';'))
			*e = 0;

		while(*p == ' ' ||
			 *p == '\t')
			p++;

		if(partial)
		{
			glong l = strlen(srch);

			if(!strncasecmp(p, srch, l))
			{
				param = g_strdup(p + l);
				break;
			}
		}
		else
		{
			if(!strcasecmp(p, srch))
			{
				param = g_strdup(p);
				break;
			}
		}

		if(e)
			p = e + 1;
		else
			break;
	}

	return param;
}

/* Content-Transfer-Encoding functions.
 */
gchar content_transfer_encoding_verify(gchar *cte)
{
     gchar *content_encodings[] = {"7bit", "8bit", "binary", "quoted-printable", "base64", NULL,};
     glong x = 0;

	g_return_val_if_fail(cte != NULL, 0);

     while(content_encodings[x])
     {
          if(stristr(cte, content_encodings[x]))
               return((gchar)x);

          x++;
     }

     return 0;
}


