/*
 * Eucalyptus
 * by Paul A. Schifferer
 *
 * Module: ops.c
 *   	 General operations functions
 *
 * Copyright  1997-9 Isengard Developments
 * All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "eucalyptus.h"

glong init_stuff(void)
{
	/* Initialize some list structures.
	 */
	list_init(&euc.Folders);	
	list_init(&euc.Composes);
	list_init(&euc.Servers);
	list_init(&euc.Dictionaries);
	list_init(&euc.Viewers);
	list_init(&euc.Taglines);
	list_init(&euc.Filters);
	list_init(&euc.Macros);
	
	list_init(&euc.addr.Addresses);
	list_init(&euc.addr.Items);
	
	/* Some mutexes.
	 */
	pthread_mutex_init(&euc.mutex.Folders, NULL);
	pthread_mutex_init(&euc.mutex.Composes, NULL);
	pthread_mutex_init(&euc.mutex.Servers, NULL);
	pthread_mutex_init(&euc.mutex.Dictionaries, NULL);
	pthread_mutex_init(&euc.mutex.Prefs, NULL);
	pthread_mutex_init(&euc.mutex.Filters, NULL);
	pthread_mutex_init(&euc.mutex.Viewers, NULL);
	
	/* Subsystems.
	 */
	smtp_init();
	pop_init();
	
	/* Message queues.
	 */
	list_init(&euc.q.Msgs);
	pthread_mutex_init(&euc.q.Mutex, NULL);

	/* Misc. things.
	 */
	
	return errno;
}

gboolean euc_startup(gboolean initial)
{
	gchar *path;
	gboolean success = TRUE;

	/* Get program information.
	 */
	 
	/* Set up directories.
	 */
	path = path_build(g_get_home_dir(), ".eucalyptus", EUCALYPTUS_FORMS_DIR, NULL);
	if(!path_create_directory(path))
	{
		g_free(path);
		return FALSE;
	}
	g_free(path);

	/* Show the about window?
	 */	 
#if 0
	if(!(euc.Flags & (EGF_QUIET | EGF_COMPOSEONLY)) ||
	   initial)
		euc_splash(-1);
#endif

	/* Set up basic set of folders.
	 */
	if(!folders_init_required())
		return FALSE;

	/* Load the folders.
	 */
	euc_read_folders();

	/* Read the address book.
	 */
	addresses_read();

	/* Read the mailcap (viewers) file.
	 */
	viewers_read();

	/* Read the filters file.
	 */
	filters_read();

	/* Open the main window.
	 */
	if(!(euc.Flags & EGF_INITCONFIG))
		euc_main_window();

	/* Set up housekeeping.
	 */
	euc.HouseKeeper = gtk_timeout_add(1 /* seconds */ * 1000,
			euc_housekeeper, 
			NULL);
	
	return success;
}

void euc_main_window(void)
{
	Folder *folder;
	GList *node;

	euc_open_window();

	folder = folder_get_special(FTF_INCOMING);
	for(node = g_list_first(euc.Folders.List); node; node = g_list_next(node))
	{
		if(folder->Data.OpenAtStartup)
		{
			folder = (Folder *)node->data;
			break;
		}
	}

	folder_open(folder);
}

void euc_shutdown(void)
{
	GList *node;
	
	/* Get rid of housekeepers and timers.
	 */
	gtk_idle_remove(euc.HouseKeeper);
	if(smtp_control.Timer)
		gtk_timeout_remove(smtp_control.Timer);
	if(pop_control.Timer)
		gtk_timeout_remove(pop_control.Timer);
	if(euc.CompTimer)
		gtk_timeout_remove(euc.CompTimer);
		
	/* Clean up threads.
	 */
	smtp_thread_stop(TRUE);
//	pop_thread_stop();
#ifdef USE_FOLDER_MONITOR
	folder_monitor_stop();
#endif
//	mailmsg_print_thread_stop();

	/* Clean up folders.
	 */
	while(node = g_list_first(euc.Folders.List))
	{
		euc.Folders.List = g_list_remove(euc.Folders.List, node->data);
		folder_free((Folder *)node->data, FALSE, FALSE);
	}


	/* Delete all temporary files.
	 */
	 
	 	
}

void euc_all_done(glong error)
{
	euc_shutdown();
	
	
	exit(error);
}

glong euc_read_folders(void)
{
	glong error = 0;
	gchar *dir;
	int fd;

	dir = path_build(g_get_home_dir(), "Mail", NULL);
	euc_scan_dir_for_folders(dir);
	
	g_free(dir);

	return error;
}

void euc_scan_dir_for_folders(gchar *dir)
{
	struct FolderData fd;
	DIR *dp;
	FILE *fp;
	gchar *fname;

	fname = path_build(dir, "FOLDER", NULL);
			
	/* Check if there's a FOLDER file here.
	 */
	if(fp = fopen(fname, "r"))
	{
		gchar buf[256];

		while(fgets(buf, sizeof(buf) - 1, fp))
		{
			gchar *c, *parm;

			if(c = strchr(buf, '\r'))
				*c = '\0';
			if(c = strchr(buf, '\n'))
				*c = '\0';

			parm = NULL;
			if(c = strchr(buf, ' '))
			{
				*c = '\0';
				parm = strip_leading_blanks(c + 1);
			}

			if(!strcmp(buf, "#"))
			{
				Folder *folder;

				fd.Flags |= FFF_UPDATEMAIN;

				if(fd.Type & FTF_SPECIAL)
				{
					folder_merge_info(&fd);
				}
				else
				{
					if(folder = folder_new(&fd))
					{
						list_append(&euc.Folders, folder);
					}
				}

				memset(&fd, 0, sizeof(struct FolderData));

				continue;
			}
			else if(!strcasecmp(buf, "Name"))
			{
				if(*parm)
					strncpy(fd.Name, parm, sizeof(fd.Name) - 1);
			}
			else if(!strcasecmp(buf, "BelongsTo"))
			{
				if(*parm)
					strncpy(fd.BelongsTo, parm, sizeof(fd.BelongsTo) - 1);
			}
			else if(!strcasecmp(buf, "Directory"))
			{
				if(*parm)
					strncpy(fd.Directory, parm, sizeof(fd.Directory) - 1);
			}
			else if(!strcasecmp(buf, "SpoolFile"))
			{
				if(*parm)
					strncpy(fd.SpoolFile, parm, sizeof(fd.SpoolFile) - 1);
			}
			else if(!strcasecmp(buf, "IMAP"))
			{
				fd.IMAP = TRUE;

				// other stuff
			}
			else if(!strcasecmp(buf, "OpenAtStartup"))
			{
				fd.OpenAtStartup = TRUE;
			}
			else if(!strcasecmp(buf, "ProtectMsgs"))
			{
				fd.ProtectMsgs = TRUE;
			}
			else if(!strcasecmp(buf, "ApplyFiltersOnImport"))
			{
				fd.ApplyFiltersOnImport = TRUE;
			}
			else if(!strcasecmp(buf, "Size"))
			{
				if(*parm)
					fd.Size = atol(parm);
			}
			else if(!strcasecmp(buf, "Sort"))
			{
				fd.SortBy = strtol(parm, NULL, 16);
			}
			else if(!strcasecmp(buf, "Flags"))
			{
				fd.Flags = strtol(parm, NULL, 16);
			}
			else if(!strcasecmp(buf, "Type"))
			{
				fd.Type = strtol(parm, NULL, 16);
			}
			else if(!strcasecmp(buf, "NextMsgNum"))
			{
				if(*parm)
					fd.NextMsgNum = atol(parm);
			}
			else if(!strcasecmp(buf, "MsgCount"))
			{
				gchar *e;

				if(c = strchr(parm, ','))
				{
					*c = '\0';
					fd.MsgCount = atol(parm);

					if(e = strchr(c + 1, ','))
					{
						*e = '\0';
						fd.NewMsgs = atol(c + 1);

						if(c = strchr(e + 1, ','))
						{
							*c = '\0';
							fd.UnreadMsgs = atol(e + 1);

							fd.NotifyCount = atol(c + 1);
						}
					}
				}
			}
			else if(!strcasecmp(buf, "MaxMsgs"))
			{
				if(*parm)
					fd.MaxMsgs = atol(parm);
			}
#ifdef EUCALYPTUS_PRO
			else if(!strcasecmp(buf, "DefaultTo"))
			{
				if(*parm)
					strncpy(fd.DefaultTo, parm, sizeof(fd.DefaultTo) - 1);
			}
#endif
		}
		
		fclose(fp);
	}
	
	/* Scan for subdirectories, and check them too.
	 */
	if(dp = opendir(dir))
	{
		struct dirent *de;
		
		while(de = readdir(dp))
		{
			gchar *dirpath;
			
			/* Skip "." directories.
			 */
			if(de->d_name[0] == '.')
				continue;

			/* Check for a dir and recursively
			 * scan.
			 */			
			dirpath = path_build(dir, de->d_name, NULL);
			if(path_is_dir(dirpath))
			{
				euc_scan_dir_for_folders(dirpath);
			}
			
			g_free(dirpath);
		}

		closedir(dp);
	}
	
	g_free(fname);
}

void euc_main_iteration(void)
{
	euc.Flags |= EGF_DISPATCHED;
	gtk_main_iteration();
	euc.Flags &= ~EGF_DISPATCHED;
}

/* Miscellaneous functions.
 */
gint euc_housekeeper(gpointer data)
{
	GList *node;
	eMsg *emsg;
	glong i;
	gboolean update = FALSE, empty;

	/* Check if we're called during a
	 * gtk_main_iteration() caused by
	 * a dispatched function.
	 */
	if(euc.Flags & (EGF_DISPATCHED | EGF_INITCONFIG))
		   return FALSE;

	/* Check for timers that need to be started.
	 * e.g., the user has changed the timeout
	 * for the POP or SMTP threads.
	 */
	if(!smtp_control.Timer && 
	   !smtp_control.th.Thread &&
	   prefs.SendInterval)
	{
		/* Start the timer.
		 */
		smtp_control.Timer = gtk_timeout_add(prefs.SendInterval * 60 * 1000,
				smtp_timer,
				NULL);
	}

	if(!pop_control.Timer && 
	   prefs.GetInterval)
	{
		/* Start the timer.
		 */
		pop_control.Timer = gtk_timeout_add(prefs.GetInterval * 60 * 1000,
				euc_pop_timer,
				NULL);
	}

	/* Check for folder updates.
	 */
	for(node = g_list_first(euc.Folders.List); node; node = g_list_next(node))
	{
		Folder *folder = (Folder *)node->data;
		
		if(folder->Data.Flags & FFF_UPDATE)
		{
			update = TRUE;

			/* FIXME */

			folder->Data.Flags &= ~FFF_UPDATE;
		}
	}
	
	/* Check for dead composes.
	 */
	i = 0;
	node = g_list_nth(euc.Composes.List, i);
	while(node)
	{
		Composition *comp = (Composition *)node->data;
		
		if(comp->Flags & CF_DONE)
		{
			euc.Composes.List = g_list_remove(euc.Composes.List, comp);
			compose_free(comp);
		}
		else
			i++;
	
		node = g_list_nth(euc.Composes.List, i);
	}

	/* Check the "message" queue.
	 */
	empty = FALSE;
	while(!empty)
	{
		emsg = NULL;

		pthread_mutex_lock(&euc.q.Mutex);
		if(node = g_list_first(euc.q.Msgs.List))
		{
			emsg = (eMsg *)node->data;
			euc.q.Msgs.List = list_remove(&euc.q.Msgs, emsg);
		}
		else
			empty = TRUE;
		pthread_mutex_unlock(&euc.q.Mutex);

		if(emsg)
		{
			   switch(emsg->type)
			   {
				   ProgressMsg *pmsg;
				   CopyMsg *cmsg;
				   GenericMsg *gmsg;
				   NewMsg *nmsg;
				   UpdateMsg *umsg;
				   StatusMsg *stmsg;
				   ErrorMsg *ermsg;
				   struct stat *st;
				   gchar *meta;

				   case MSG_TYPE_ERROR:
					   ermsg = (ErrorMsg *)emsg;
					   {
							 gchar *str;

							 str = g_strdup_printf("%s\nError %d: %s",
										  ermsg->text,
										  ermsg->error, strerror(ermsg->error));
							 euc_request(euc.Window,
										  _("Eucalyptus error"),
										  str,
										  euc_ok_button, 0,
										  EUCRF_ASYNC);
							 g_free(str);
					   }
					   break;
				   
				   case MSG_TYPE_NEW:
						nmsg = (NewMsg *)emsg;
						
						st = NULL;
						meta = NULL;
						if(st = path_stat(nmsg->filename))
						{
							gint size;

						   	gnome_metadata_get(nmsg->filename, "msg-attrs", &size, &meta);
							if(folder_process_new_msg(nmsg->folder, nmsg->filename, st->st_size, meta))
								update = TRUE;
						}

						g_free(st);
						g_free(meta);
					   break;
					   
				   case MSG_TYPE_PROGRESS:
					   pmsg = (ProgressMsg *)emsg;
					   
					   switch(pmsg->who)
					   {
						   case MSG_WHO_SMTP:
							   switch(pmsg->msg.subtype)
							   {
								   case MSG_SUBTYPE_PROGRESS_OPEN:
								        smtp_open_window();
									   break;
									    
								   case MSG_SUBTYPE_PROGRESS_CLOSE:
									   smtp_close_window();
									   break;

								   default:
									   smtp_update_progress(pmsg);
									   break;
							   }
							   break;
							   
						   case MSG_WHO_POP:
							   switch(pmsg->msg.subtype)
							   {
								   case MSG_SUBTYPE_PROGRESS_OPEN:
								        pop_open_window(pmsg->server);
									   break;
									    
								   case MSG_SUBTYPE_PROGRESS_CLOSE:
									   pop_close_window(pmsg->server);
									   break;

								   default:
									   pop_update_progress(pmsg->server, pmsg);
									   break;
							   }

							   break;
					   }
					   break;
					   
				   case MSG_TYPE_COPY:
					   {
						   gchar filename[FILENAME_MAX];
						   
						   cmsg = (CopyMsg *)emsg;
						   
						   /* Copy the message file.
						    */
						   if(mailmsg_copy(cmsg->filename, cmsg->dest, filename, sizeof(filename)))
						   {
							   update = TRUE;
							   
							   /* Create an index entry for the 
							    * destination folder.
							    */
							   folder_update_index(cmsg->dest);
							   folder_update_visual(cmsg->dest);

							   /* Check if it should be moved.
							    */
							   if(cmsg->move)
							   {
								   /* Remove the entry.
								    */
								   mailmsg_remove(NULL, cmsg->filename);
								   
								   /* Remove the index from the source
								    * folder.
								    */
								   folder_update_index(cmsg->source);
								   folder_update_visual(cmsg->source);
							   }
						   }
						   else
						   {
							   gchar *msg = g_strdup_printf(_("Unable to copy/move message '%s'\nfrom '%s' to '%s':\n%s"),
									   cmsg->filename, cmsg->source->Data.Name,
									   cmsg->dest->Data.Name,
									   strerror(errno));

							   euc_request(euc.Window,
									   _("Eucalyptus message"),
									   msg,
									   euc_ok_button,
									   0, EUCRF_ASYNC);
							   g_free(msg);
						   }
					   }
					   break;
					   
				   case MSG_TYPE_GENERIC:
					   break;
					   
				   case MSG_TYPE_UPDATE:
					   umsg = (UpdateMsg *)emsg;
					   
					   folder_update_index(umsg->folder);
					   folder_update_visual(umsg->folder);
					   update = TRUE;
					   break;

				case MSG_TYPE_STATUS:
					stmsg = (StatusMsg *)emsg;

					break;
			}

			g_free(emsg);
		}
	}

	if(update)
	{
		euc_update_listview();
	}

	return TRUE;
}

gint euc_pop_timer(gpointer data)
{
}

/* Callbacks.
 */
gint euc_quit_requested_cb(GtkWidget *window, gpointer data)
{
	GtkWidget *dialog;

	if(prefs.AskBeforeQuitting)
	{
		if(euc_request(window, 
				_("Quit Eucalyptus"),
				_("Are you sure you want to quit?"),
				euc_ok_cancel_buttons, 1,
				0))
			return TRUE;
	}

	gtk_main_quit();

	return FALSE;
}

