/*
 * Eucalyptus
 * by Paul A. Schifferer
 *
 * Module: pop.c
 *   	 POP client function server
 *
 * Copyright  1997-9 Isengard Developments
 * All rights reserved.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "eucalyptus.h"
#include "md5.h"

struct POPControl pop_control;

/* Servers.
 */
Server *pop_server_new(gchar type)
{
	Server *server = NULL;
	
	if(server = g_malloc0(sizeof(Server)))
	{
		   list_init(&server->Messages);

		   switch(type)
		   {
				 case PST_POP3:
					    strncpy(server->Host, "pop", sizeof(server->Host) - 1);
					    server->Port = 110;
					    break;
				 case PST_IMAP4:
					    strncpy(server->Host, "imap", sizeof(server->Host) - 1);
					    server->Port = 153;
					    break;
		   }

		   server->Socket = -1;
	}
	
	return server;
}

void pop_server_free(Server *server)
{
	   GList *node;

	if(!server)
		return;

	while(node = g_list_first(server->Messages.List))
	{
	}

	g_free(server);
}

void pop_server_generate_id(Server *server)
{
	MD5_CTX context;
	glong i;
	gchar digest[16], *id;

	g_return_if_fail(server != NULL);

	id = g_strdup_printf("%s%s", server->Host, server->Username);
	
	MD5Init(&context);
	MD5Update(&context, id, strlen(id));
	MD5Final(digest, &context);
	
	g_free(id);

	strcpy(server->ID, "");
	for(i=0; i<16; i++)
	{
		gchar hex[5];

		sprintf(hex, "%02x", digest[i]);
		strncat(server->ID, hex, sizeof(server->ID) - 1);
	}
}

Server *pop_server_lookup_id(GList *list, gchar *id)
{
	GList *node;
	
	for(node = g_list_first(list); node; node = g_list_next(node))
	{
		Server *s;
		
		s = (Server *)node->data;
		
		if(!strcmp(id, s->ID))
			return s;
	}
	
	return NULL;
}

gboolean pop_server_connect(Server *server)
{
	if(!server)
		return FALSE;

	server->Flags &= ~PISF_BADDISCONNECT;

	if((server->Socket < 0) &&
	   !server->Disabled)
	{
		struct hostent *host;

		pop_server_reset(server);
		pop_server_status(server, _("Connecting..."), 0);

		server->SockAddr.sin_family = AF_INET;
		server->SockAddr.sin_port = htons(server->Port);
		server->SockAddr.sin_addr.s_addr = 0;

		if(host = gethostbyname(server->Host))
		{
			memcpy(&server->SockAddr.sin_addr, host->h_addr, host->h_length);

			if(-1 != (server->Socket = socket(server->SockAddr.sin_family, SOCK_STREAM, 0)))
			{
				if(0 == connect(server->Socket, (struct sockaddr *)&server->SockAddr, sizeof(server->SockAddr)))
				{
					server->Connected = TRUE;
				}
				else
				{
				}
			}
		}
	}

	// now open the server and do the initial greet and check
	if((server->Socket >= 0) &&
	   server->Connected)
	{
#if 0 // FIXME
		// if no password, ask for one
		if(!strlen(server->Password) &&
		   !strlen(server->SessionPass))
		{
			gchar prompt[201];

			sprintf(prompt, LocaleString(ETXT_POP_INPUTPASSgint16), server->Host);
			if(!RequestString(prompt, server->SessionPass, sizeof(server->SessionPass), RSF_PASSgint16))
			{
				return(FALSE);
			}
		}
		else
#endif
		strncpy(server->SessionPass, server->Password, sizeof(server->SessionPass) - 1);

		// wait for greeting
		if(0 == pop_server_send(server, "", server->in_buf, sizeof(server->in_buf)))
		{
			glong val;
			gchar *c;

			pop_server_status(server, _("Connecting..."), 0);

			if(c = strchr(server->in_buf, '<'))
			{
				gchar *e;

				if(e = strchr(c, '>'))
				{
					MD5_CTX context;
					glong i;
					gchar digest[16], dstring[101];

					*(e + 1) = '\0';

					strncpy(dstring, c, sizeof(dstring) - 1);
					strncat(dstring, server->SessionPass, sizeof(dstring) - 1);

					MD5Init(&context);
					MD5Update(&context, dstring, strlen(dstring));
					MD5Final(digest, &context);

					strcpy(dstring, "");
					for(i=0; i<16; i++)
					{
						gchar hex[5];

						sprintf(hex, "%02x", digest[i]);
						strncat(dstring, hex, sizeof(dstring) - 1);
					}

					pop_server_status(server, _("Authenticating..."), 0);

					sprintf(server->out_buf, "APOP %s %s\r\n", server->Username, dstring);
					switch(val = pop_server_send(server, server->out_buf, server->in_buf, sizeof(server->in_buf)))
					{
						case 0:   // good
							pop_server_status(server, _("Authenticated."), 0);
							server->Authenticated = TRUE;
							break;
						default:
							pop_server_disconnect(server);
							pop_server_status(server, _("Bad authentication."), 0);
							server->Flags |= PISF_BADDISCONNECT;
							server->State = PSS_STOPPING;
							break;
					}

					goto auth_out;
				}
			}

			sprintf(server->out_buf, "USER %s\r\n", server->Username);
			switch(val = pop_server_send(server, server->out_buf, server->in_buf, sizeof(server->in_buf)))
			{
				case 0:   // good
					pop_server_status(server, _("Authenticating..."), 0);

					sprintf(server->out_buf, "PASS %s\r\n", server->SessionPass);
					switch(val = pop_server_send(server, server->out_buf, server->in_buf, sizeof(server->in_buf)))
					{
						case 0:   // good
							pop_server_status(server, _("Authenticated."), 0);
							server->Authenticated = TRUE;
							break;
						default:
							pop_server_disconnect(server);
							pop_server_status(server, _("Bad authentication."), 0);
							server->Flags |= PISF_BADDISCONNECT;
							server->State = PSS_STOPPING;
							break;
					}
					break;
				default:
					pop_server_disconnect(server);
					pop_server_status(server, _("Bad authentication."), 0);
					server->Flags |= PISF_BADDISCONNECT;
					server->State = PSS_STOPPING;
					break;
			}

			auth_out:
			if(server->Authenticated)
			{
				switch(val = pop_server_send(server, "STAT\r\n", server->in_buf, sizeof(server->in_buf)))
				{
					gchar *c;

					case 0:
						server->Count = atol(server->in_buf);
						server->Current = 1;
						if(c = strchr(server->in_buf, ' '))
							server->Size = atol(strip_leading_blanks(c + 1));
fprintf(stderr, "server->Size = %d\n", server->Size);
						break;
					default:
						pop_server_disconnect(server);
						pop_server_status(server, server->in_buf, 0);
						server->Flags |= PISF_BADDISCONNECT;
						server->State = PSS_STOPPING;
						break;
				}
			}
		}
	}

	return server->Connected;
}

void pop_server_disconnect(Server *server)
{
	if(!server)
		return;

	if(server->Connected)
	{
		send(server->Socket, "QUIT\r\n", 6, 0);
	}

	if(server->Socket > -1)
	{
		shutdown(server->Socket, 2);
	}
	server->Socket = -1;

	server->Connected = FALSE;
	server->Authenticated = FALSE;

	pop_server_reset(server);

	if(!(server->Flags & PISF_BADDISCONNECT))
		pop_server_status(server, _("Not connected."), 0);
}

glong pop_server_send(Server *server, gchar *out_buf, gchar *in_buf, glong in_len)
{
	glong error = 0;

	if(!server)
		return(-1);

	if(-1 != send(server->Socket, out_buf, strlen(out_buf), 0))
	{
		gchar *c;

		memset(in_buf, 0, in_len);

		if(pop_read_line(server->Socket, in_buf, in_len - 1) < 1)
			return(-1);
		if(!strncasecmp(in_buf, "+OK", 3))
			error = 0;
		if(!strncasecmp(in_buf, "-ERR", 4))
			error = 1;

		if(c = strchr(in_buf, ' '))
			strcpy(in_buf, c + 1);
	}
	else
		error = -1;

	return(error);
}

void pop_server_reset(Server *server)
{
	if(!server)
		return;

	server->Socket  		= -1;
	server->Connected    	=
	server->Authenticated	= FALSE;
	server->Count   		=
	server->Current 		= 0;
	server->NOOPs   		= 0;
	server->Flags   	    &= ~(PISF_NOMOREMSGS);
	server->Size    		= 1;
fprintf(stderr, "server->Size = %d\n", server->Size);
	server->Level   		= 0;

	memset(&server->SockAddr, 0, sizeof(server->SockAddr));

//	server->SockAddr.sin_len    	= sizeof(server->SockAddr);
	server->SockAddr.sin_family 	= AF_INET;
	server->SockAddr.sin_port   	= server->Port;
	server->SockAddr.sin_addr.s_addr = 0;
}


