/*
 * Eucalyptus
 * by Paul A. Schifferer
 *
 * Module: euc_request.c
 *         Customized dialog functions
 *
 * Copyright (c) 1997-9 Isengard Developments
 * All rights reserved.
 */

#include <gnome.h>

#include "eucalyptus.h"

const gchar *euc_ok_cancel_buttons[] = { GNOME_STOCK_BUTTON_OK, GNOME_STOCK_BUTTON_CANCEL, NULL, };
const gchar *euc_yes_no_buttons[] = { GNOME_STOCK_BUTTON_YES, GNOME_STOCK_BUTTON_NO, NULL, };
const gchar *euc_ok_button[] = { GNOME_STOCK_BUTTON_OK, NULL, };

typedef struct _EucRequest EucRequest;
struct _EucRequest
{
	GtkWidget *request,
			*label;
	gchar *title,
		 *text;
};

gint euc_request(GtkWidget *parent, const gchar *title, const gchar *text, const gchar **buttons, gint def, gint16 flags)
{
	EucRequest *ereq;
	GtkWidget *label;
	gint answer = 0;

	g_return_val_if_fail(parent != NULL, def);
	g_return_val_if_fail(text != NULL, def);
	
	ereq = g_malloc0(sizeof(EucRequest));
	if(!ereq)
		return def;

	ereq->request = gnome_dialog_newv(ereq->title = title ? g_strdup(title) : g_strdup("Eucalyptus"), 
						   buttons ? buttons : euc_ok_button);

	gnome_dialog_set_close(GNOME_DIALOG(ereq->request), TRUE);
	gnome_dialog_close_hides(GNOME_DIALOG(ereq->request), FALSE);

	if(parent && !(flags & EUCRF_ASYNC))
		gnome_dialog_set_parent(GNOME_DIALOG(ereq->request), GTK_WINDOW(parent));
	
	gnome_dialog_set_default(GNOME_DIALOG(ereq->request), def);

	label = gtk_label_new(ereq->text = g_strdup(text));
	gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(ereq->request)->vbox), label, TRUE, TRUE, GNOME_PAD_SMALL);
	
	gtk_widget_show_all(ereq->request);

	if(flags & EUCRF_ASYNC)
	{
		/* Set a callback to destroy this dialog
		 * when the user clicks okay.
		 */
		gnome_dialog_button_connect(GNOME_DIALOG(ereq->request),
				0, GTK_SIGNAL_FUNC(euc_request_destroy), ereq);
	}
	else
	{
		/* Run the dialog.
		 */
		answer = gnome_dialog_run_and_close(GNOME_DIALOG(ereq->request));

		g_free(ereq->title);
		g_free(ereq->text);
	
		g_free(ereq);
	}

	return answer;
}

void euc_request_destroy(GtkButton *button, gpointer data)
{
	EucRequest *ereq = (EucRequest *)data;
	
	gtk_widget_destroy(ereq->request);
	
	g_free(ereq->title);
	g_free(ereq->text);
	
	g_free(ereq);
}

/* File requester functions.
 */
FileReq *euc_file_request_new(glong type, gchar *title, gchar *filename)
{
	FileReq *req;
	
	/* Create the requester structure.
	 */
	req = g_malloc0(sizeof(FileReq));
	
	/* Create the GtkFileSelection.
	 */
	req->requester = gtk_file_selection_new(title);
	req->type = type;

	/* Set the forms dir.
	 */
	gtk_file_selection_set_filename(GTK_FILE_SELECTION(req->requester), filename);

	/* Connect signals.
	 */
	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(req->requester)->ok_button), "clicked",
			GTK_SIGNAL_FUNC(euc_file_request_okay_cb), req);
	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(req->requester)->cancel_button), "clicked",
			GTK_SIGNAL_FUNC(euc_file_request_cancel_cb), req);
	gtk_signal_connect(GTK_OBJECT(req->requester), "destroy",
			GTK_SIGNAL_FUNC(euc_file_request_destroy_cb), req);

	/* Show it.
	 */	
	gtk_widget_show_all(req->requester);

	return req;
}

/* File requester callbacks.
 */
void euc_file_request_okay_cb(GtkWidget *button, gpointer data)
{
	FileReq *req = (FileReq *)data;
	gchar *fname;

	/* Get the filename/dir.
	 */
	fname = gtk_file_selection_get_filename(GTK_FILE_SELECTION(req->requester));
	
	/* Handle the requester according
	 * to type.
	 */
	switch(req->type)
	{
		Composition *comp;
		
		case FILEREQ_COMPOSE_FORM:
			if(!path_is_file(fname))
				return;
				
			comp = compose_new();
			list_append(&euc.Composes, comp);
			compose_open_window(comp);
			compose_insert_form(comp, fname);
			break;
	}
	
	gtk_widget_unref(req->requester);
}

void euc_file_request_cancel_cb(GtkWidget *button, gpointer data)
{
	FileReq *req = (FileReq *)data;

	gtk_widget_unref(req->requester);
}

void euc_file_request_destroy_cb(GtkWidget *button, gpointer data)
{
	FileReq *req = (FileReq *)data;
	
	/* Clean up the requester data.
	 */	
	g_free(req);
}
